import os
from pathlib import Path

from socialgene.neo4j.neo4j_element import Node
from socialgene.neo4j.schema.socialgene_modules import SocialgeneModules
from socialgene.utils.logging import log
from socialgene.utils.run_subprocess import run_subprocess


class Neo4jAdminImport(SocialgeneModules):
    """Socialgene's Neo4jAdminImport class handles building the Docker/Neo4j admin import command line arguments"""

    def __init__(
        self,
        neo4j_top_dir: str,
        cpus: int = 1,
        additional_args: str = "",
        uid: int = None,
        gid: int = None,
        module_list: list = None,
        dbms_connector_http_listen_address: int = 7474,
        dbms_connector_bolt_listen_address: int = 7687,
        docker_version="latest",
        *args,
        **kwargs,
    ):
        """Class for building and executing Neo4j admin import

        Args:
            neo4j_top_dir (str): directory containing the necessary dirs and files for Neo4j admin import
            module_list (list): name of grouped nodes/relationships (socialgene/neo4j/schema/define_modules.py)
            hmm_list (list): list of hmms source databases used (socialgene/neo4j/schema/define_hmmlist.py)
            cpus (int, optional): [description]. Defaults to 1.
            additional_args (str, optional): [description]. Defaults to "".
            uid (int, optional): [description]. Defaults to None.
            gid (int, optional): [description]. Defaults to None.
        """
        super().__init__(*args, **kwargs)
        self.input_sg_modules = module_list
        self.neo4j_top_dir = neo4j_top_dir
        self.cpus = cpus
        self.additional_args = additional_args
        self.node_relationship_argument_list = []
        self.dbms_connector_http_listen_address = dbms_connector_http_listen_address
        self.dbms_connector_bolt_listen_address = dbms_connector_bolt_listen_address
        self.docker_version = docker_version
        self.filepaths_to_check = []
        # UID/GID for docker call
        if uid is None:
            self.uid = "$(id -u)"
        else:
            self.uid = uid
        if gid is None:
            self.gid = "$(id -g)"
        else:
            self.gid = gid
        self.add_modules(module_list)

    @staticmethod
    def create_neo4j_directories(neo4j_top_dir):
        """Check Neo4j import directories, if they don't exist, create them or Neo4j's Docker container will create them as root user

        Args:
            neo4j_top_dir (str): directory to make subdirectories in, directory should already contain the Neo4j 'import' directory containing data to be imported into the database

        Raises:
            IOError: _description_
            IOError: _description_
        """
        dirs_to_check = ["data", "logs", "plugins"]
        for single_dir in dirs_to_check:
            dir_to_check = os.path.join(neo4j_top_dir, single_dir)
            if not os.path.exists(dir_to_check):
                # create directory if it doesn't exist, or neo4j will get mad
                os.mkdir(dir_to_check)
                log.info(f"Created directory for Neo4j admin import: {dir_to_check}")
        # plugins doesn't have to be empty (and is actually filled in the nextflow workflow)
        # dirs_to_check = ["data", "logs"]
        # for single_dir in dirs_to_check:
        #     dir_to_check = os.path.join(neo4j_top_dir, single_dir)
        #     if len(os.listdir(dir_to_check)) > 0:
        #         # make sure the directory is empty
        #         log.exception(
        #             f"Directory ({dir_to_check}) must be empty for Neo4j admin import to work"
        #         )
        #         raise IOError
        files_to_check = ["import.report"]
        for single_file in files_to_check:
            file_to_check = os.path.join(neo4j_top_dir, single_file)
            if not os.path.exists(file_to_check):
                # create file if it doesn't exist, or neo4j will get mad
                with open(file_to_check, "w"):
                    pass
                log.info(f"Created file for Neo4j admin import: {file_to_check}")
            if os.path.getsize(file_to_check) > 0:
                # make sure the file is empty
                raise IOError(
                    f"File ({file_to_check}) must be empty for Neo4j admin import to work"
                )

    # while building arguments, make a list of data filepaths to check for
    @staticmethod
    def _single_arg_string_builder(input, type):
        """Builds the command line arguments to import data into a new Neo4j database

        Args:
            input: node/rel class
            type(str): nodes/relationships

        Returns:
            list: [first_part_of_arg_string, header_path_string, data_glob_string] want mutable because will check in later step for gz and append if needed
        """
        cli_label = ""
        if issubclass(input, Node):
            if len(input.neo4j_label) > 1:
                # labels are set in-file
                cli_label = ""
            else:
                cli_label = f"{input.neo4j_label[0]}="
        else:
            cli_label = f"{input.neo4j_label}="
        return [
            f"--{type}={cli_label}",
            f"import/neo4j_headers/{input.header_filename}",
            f"import/{input.target_subdirectory}/*.{input.target_extension}.*",
        ]

    def _escape_arg_glob(self):
        for i in self.node_relationship_argument_list:
            i[2] = i[2].replace("*.", ".*\\.")

    def build_nodes_and_relationships_argument_list(self):
        for node in self.selected_nodes:
            self.node_relationship_argument_list.append(
                self._single_arg_string_builder(input=node, type="nodes")
            )
        for rel in self.selected_relationships:
            self.node_relationship_argument_list.append(
                self._single_arg_string_builder(input=rel, type="relationships")
            )

    def _check_files(self):
        missing_input_headers = []
        missing_input_data = []
        p = Path(self.neo4j_top_dir)
        for i in self.node_relationship_argument_list:
            if not list(p.glob(i[1])):
                missing_input_headers.append(i[1])
            if not p.glob(i[2]):
                missing_input_data.append(i[2])
        if missing_input_headers or missing_input_data:
            if missing_input_headers:
                message = f"Couldn't find expected header files for neo4j admin import:\n{missing_input_headers}"
            elif missing_input_data:
                message = f"Couldn't find expected data files for neo4j admin import:\n{missing_input_data}"
            else:
                message = "??"
            raise ValueError(message)
        else:
            log.info("Found all expected input files. Huzzah!")

    def _neo4j_admin_import_args(self, docker=False):
        """Function used to create a list of all the commands/arguments for a docker/neo4j admin import run

        Returns:
            list: commands/arguments to pass to a subprocess
        """

        docker_command = [
            "docker",
            "run",
            "--rm",
            "--interactive",
            f"--publish={self.dbms_connector_http_listen_address}:{self.dbms_connector_http_listen_address}",
            f"--publish={self.dbms_connector_bolt_listen_address}:{self.dbms_connector_bolt_listen_address}",
            f"--volume={self.neo4j_top_dir}/data:/opt/conda/bin/neo4j/data",
            f"--volume={self.neo4j_top_dir}/import:/opt/conda/bin/neo4j/import",
            f"--volume={self.neo4j_top_dir}/plugins:/opt/conda/bin/neo4j/plugins",
            f"--volume={self.neo4j_top_dir}/import.report:/opt/conda/bin/neo4j/import.report",
            f"--user={self.uid}:{self.gid}",
            f"chasemc2/sgnf-sgpy:{self.docker_version}",
        ]

        neo4j_admin_command = ["neo4j-admin", "database", "import", "full"]

        # self.node_relationship_argument_list will be inserted between pre and post
        joined_args = [
            f"{i[0]}{i[1]},{i[2]}" for i in self.node_relationship_argument_list
        ]
        post = [
            '--delimiter="\\t"',
            "--high-parallel-io=On",
            f"--threads={self.cpus}",
            "--ignore-empty-strings=true",
            "--ignore-extra-columns=true",
            "--skip-bad-relationships",
            "--skip-duplicate-nodes",
            "--normalize-types=false",
            "neo4j",  # this is the database name. If I remember correctly this must be "neo4j" for the community edition of neo4j
        ]

        if docker:
            pre = docker_command + neo4j_admin_command
        else:
            pre = neo4j_admin_command
        return pre + list(set(joined_args)) + post

    @staticmethod
    def _launch_sub(cmd):
        run_subprocess(
            command_list=" ".join(cmd),
            check=False,
            shell=True,
            capture_output=True,
        )

    def _prerun(self):
        self.create_neo4j_directories(self.neo4j_top_dir)
        self.build_nodes_and_relationships_argument_list()
        self._escape_arg_glob()

    def run(self, docker=False, dryrun=False):
        """Run the created commands/arguments in a separate process"""
        self._prerun()
        cmd_arg_list = self._neo4j_admin_import_args(docker)
        if dryrun:
            return cmd_arg_list
        else:
            self._check_files()
            self._launch_sub(cmd_arg_list)



# from socialgene.neo4j.neo4j import GraphDriver
# import logging
# import os
# from socialgene.base.chem import ChemicalCompound
# from rich.progress import Progress
# from socialgene.addons.chemistry.nr import ChemicalSubstructure, ContainsSubstructure, ChemicalCompoundNode
# from rdkit import Chem
# from multiprocessing import Pool
# from rdkit.Chem.MolStandardize import rdMolStandardize
# from itertools import batched
# logging.getLogger("neo4j").setLevel(logging.WARNING)
# logging.getLogger().setLevel(logging.INFO)

# if __name__ == "__main__":

#     #inspect(a)
#     with GraphDriver() as db:
#         results = db.run(
#             """
#             MATCH (n:chemical_compound) RETURN n.inchi as inchi
#             """
#         ).value()


#     nodes=set()
#     rels=set()

#     def process_subgraph(batched_results):
#         nodes=set()
#         rels=set()
#         for i in batched_results:
#             sgmol = ChemicalCompound(i)
#             subgraphs = Chem.FindAllSubgraphsOfLengthN(sgmol.mol, 5)
#             for subgraph in subgraphs:
#                 sub_mol = Chem.PathToSubmol(sgmol.mol, subgraph, useQuery=True)
#                 sub_mol = rdMolStandardize.Cleanup(sub_mol)
#                 sub_mol = rdMolStandardize.Normalize(sub_mol)
#                 r = rdMolStandardize.Reionizer()
#                 sub_mol = r.reionize(sub_mol)
#                 Chem.RemoveStereochemistry( sub_mol )
#                 #sub_mol=Chem.MolToSmiles(sub_mol, canonical=True)
#                 node = ChemicalSubstructure()
#                 temp_compound = ChemicalCompound(sub_mol)
#                 node.fill_from_dict(temp_compound.base_properties | temp_compound.hash_dict)
#                 del temp_compound
#                 nodes.add(node)
#                 cn = ChemicalCompoundNode()
#                 cn.fill_from_dict(sgmol.base_properties | sgmol.hash_dict)
#                 rel = ContainsSubstructure(cn, node)
#                 rels.add(rel)
#         nodes[0].add_multiple_to_neo4j(nodes)
#         rels[0].add_multiple_to_neo4j(rels)

#     def update_progress(*args):
#         pg.update(task, advance=1)

#     with Progress(transient=True) as pg:
#         task = pg.add_task("Progress...", total=len(results))
#         batched_results = batched(results, 100)
#         with Pool() as p:
#             for i in p.imap_unordered(process_subgraph, batched_results):
#                 nodes.update(i[0])
#                 rels.update(i[1])
#                 update_progress()



import csv
import os
import re
from collections import OrderedDict, defaultdict
from dataclasses import dataclass, field
from pathlib import Path, PureWindowsPath
from posixpath import normpath
from typing import List

import socialgene.utils.file_handling as fh
from socialgene.hashing.hashing import hasher
from socialgene.utils.logging import log

re_pfam_broad = re.compile("^PF[0-9]{5,5}")


HMM_SOURCES = [
    "amrfinder",
    "antismash",
    "bigslice",
    "classiphage",
    "ipresto",
    "local",
    "pfam",
    "prism",
    "resfams",
    "tigrfam",
    "virus_orthologous_groups",
]


def parse_hmmlist_input(input):
    # Filter hmm databases based on input list of hmm database names or "all"
    # accept "all" as a list or string
    if isinstance(input, str):
        input = [input]
    if "all" in input:
        _hmms = HMM_SOURCES
    else:
        _hmms = [i for i in input if i in HMM_SOURCES]
    return _hmms


@dataclass
class HmmModel:
    """
    The above code defines a data class called `HmmModel` in Python. This class represents a model for a
    Hidden Markov Model (HMM). It has various attributes that store information about the model, such as
    its name, description, length, maximum length, alphabet, and other properties.
    """

    _n: int = None
    _base_dir: str = None
    _abs_path: str = None
    _rel_path: str = None
    _model_source: str = None
    _super_category: str = None
    _category: str = None
    _subcategory: str = None
    _hash: str = None
    _new_hash: str = None
    _pfam_accession: str = None
    _pfam_version: int = 0
    _notes: List[str] = field(default_factory=lambda: [])
    _unknown: List[str] = field(default_factory=lambda: [])
    HMMER3_f: str = None
    NAME: str = None
    ACC: str = None
    DESC: str = None
    LENG: str = None
    MAXL: str = None
    ALPH: str = None
    RF: str = None
    MM: str = None
    CONS: str = None
    CS: str = None
    MAP: str = None
    DATE: str = None
    COM: str = None
    NSEQ: str = None
    EFFN: str = None
    CKSUM: str = None
    GA: str = None
    TC: str = None
    NC: str = None
    STATS: List[str] = field(default_factory=lambda: [])
    HMM: str = None
    COMPO: str = None
    MODEL: List[str] = field(default_factory=lambda: [])

    @property
    def has_cutoffs(self):
        if self.GA:
            return True
        else:
            return False

    def all_attributes(self):
        """Template dictionary to hold the desired info of a single hmm model"""
        if isinstance(self._notes, list):
            notes = "; ".join(self._notes)
        else:
            notes = str(self._notes)

        if self._new_hash:
            hash_to_use = self._new_hash
        else:
            hash_to_use = self._hash
        # need tigrfam ids to actually be the tigrfam ids (not f"{self._model_source}_{self._n}"),
        # so that tigrfam_to_go, etc will associate
        # need others to be f"{self._model_source}_{self._n}" because can't assume they will be
        # unique within a source db (ie same model in different antismash categories)
        if self._model_source == "tigrfam":
            mod_id = self.NAME
        else:
            mod_id = f"{self._model_source}_{self._n}"
        return OrderedDict(
            {
                "id": mod_id,
                "source": self._model_source,
                "rel_path": self._rel_path,
                "name": self.NAME,
                "acc": self.ACC,
                "notes": notes,
                "description": self.DESC,
                "date": self.DATE,
                "hash": self._hash,
                "hash_used": hash_to_use,
                "model_length": len(self.MODEL),
                "super_category": self._super_category,
                "category": self._category,
                "subcategory": self._subcategory,
                "ga": self.GA,
                "tc": self.TC,
                "nc": self.NC,
            }
        )

    def add_attr(self, line):
        """
        The function `add_attr` checks if a line is an HMM attribute and adds it to the appropriate
        attribute in the class.

        Args:
          line: The `line` parameter is a string of text coming from a file parse/read. It represents a
        single line of text that is being processed by the `add_attr` method.
        """
        stripped_line = line.strip()
        line_val = stripped_line.split(maxsplit=1)
        line_val = [i.rstrip() for i in line_val]
        if not line_val:
            return
        if stripped_line.startswith("HMMER3"):
            self.HMMER3_f = stripped_line
        elif line_val[0] == "STATS":
            self.STATS.append(line_val[1])
        elif line_val[0] in self.__dataclass_fields__.keys():
            setattr(self, line_val[0], line_val[1])
        else:
            self._unknown.append(stripped_line)

    def add_model(self, line):
        self.MODEL.append(line.rstrip())

    def add_model_hash(self):
        self._hash = hasher("".join(self.MODEL))
        self._new_hash = self._hash

    def find_pfam_accessions(self):
        """
        Attempts to extract Pfam accessions and versions from a given ACC string and assigns them to instance variables `_pfam_accession` and `_pfam_version`
        respectively.
        """
        try:
            if re_pfam_broad.match(self.ACC):
                temp = self.ACC.split(".", maxsplit=1)
                self._pfam_accession = temp[0]
                self._pfam_version = int(temp[1])
        except Exception:
            # no error, just don't assign pfam vars
            pass

    def _get_gen_attr_str(self, attr):
        """
        The function returns a formatted string representation of a given attribute if it is not empty.

        Args:
          attr: The `attr` parameter is a string representing the name of an attribute.

        Returns:
          a formatted string that includes the attribute name and its corresponding value. The attribute
        name is aligned with the string "STATS" by adding spaces before the attribute name.
        """
        if getattr(self, attr):
            n_spaces_offset = len("STATS") + 1 - len(attr)
            return f"{attr}{' ' * n_spaces_offset}{getattr(self, attr)}"

    def _get_name_as_hash(self):
        """
        The function returns the name attribute of an object along with its corresponding hash value, space aligned.

        Returns:
          a string that combines the value of the attribute "NAME" with the value of the "_new_hash"
        attribute, separated by a number of spaces.
        """
        attr = "NAME"
        n_spaces_offset = len("STATS") + 1 - len(attr)
        return f"{attr}{' ' * n_spaces_offset}{getattr(self, '_new_hash')}"

    def model_string(self, hash_as_name=False):
        """
        The `model_string` function generates a string representation of a model, including various
        attributes and headers.

        Args:
          hash_as_name: The `hash_as_name` parameter is a boolean flag that determines whether the
        model's name should be represented as a hash value or as a regular string. If `hash_as_name` is
        set to `True`, the model's name will be converted to a hash value. Defaults to
        False

        Returns:
          a string that represents the model.
        """
        res = []
        # this is spread across multiple calls because likely to include
        # intermediate if/else statemets later
        res.append(getattr(self, "HMMER3_f"))
        # write hash if...
        if hash_as_name:
            res.append(self._get_name_as_hash())
        else:
            res.append(self._get_gen_attr_str("NAME"))
        for i in [
            "ACC",
            "DESC",
            "LENG",
            "MAXL",
            "ALPH",
            "RF",
            "MM",
            "CONS",
            "CS",
            "MAP",
            "DATE",
            "COM",
            "NSEQ",
            "EFFN",
            "CKSUM",
            "GA",
            "TC",
            "NC",
        ]:
            # not all attr are required, check if they exist
            if hasattr(self, i) and getattr(self, i):
                res.append(self._get_gen_attr_str(i))
        for i in self.STATS:
            res.append(f"STATS {i}")
        # write any unknown headers that were present in the original model description
        if getattr(self, "_unknown"):
            for i in getattr(self, "_unknown"):
                res.append(i)
        # write model
        for i in self.MODEL:
            res.append(i)
        res.append("//")
        return "\n".join([i for i in res])

    def write(self, outpath, mode, hash_as_name=False):
        with open(outpath, mode=mode) as h:
            h.write(self.model_string(hash_as_name=hash_as_name))
            h.write("\n")

    def _assign_relative_path(self):
        input_dir = Path(self._base_dir)
        input_path = Path(self._abs_path)
        self._rel_path = input_path.relative_to(input_dir)

    def assign_category(self):
        """
        The function `assign_category` is used to parse categories using the input models' directory structure.

        """
        self._antismash_categories()

    def _antismash_categories(self):
        """
        The function extracts and defines categories of antismash models based on the antismash
        directory structure.

        Args:
          hmmpath: The absolute path to the hmm file.
        """
        # standardize cross-os path
        st_hmmpath = PureWindowsPath(
            normpath(PureWindowsPath(self._abs_path).as_posix())
        ).as_posix()
        if "antismash/modules" in st_hmmpath:
            self._super_category = "modules"
        if "antismash/detection" in st_hmmpath:
            self._super_category = "detection"
        if "antismash/modules/nrps_pks" in st_hmmpath:
            self._category = "nrps_pks"
            if "antismash/modules/nrps_pks/minowa/data/CAL_HMMs" in st_hmmpath:
                self._subcategory = "cal_hmms"
            if "antismash/modules/nrps_pks/minowa/data/AT_HMMs" in st_hmmpath:
                self._subcategory = "at_hmms"
        if "antismash/modules/thiopeptides" in st_hmmpath:
            self._category = "thiopeptides"
        if "antismash/modules/lanthipeptides" in st_hmmpath:
            self._category = "lanthipeptides"
            if "antismash/modules/lanthipeptides/data/non_biosyn_hmms" in st_hmmpath:
                self._subcategory = "non_biosyn_hmms"
        if "antismash/modules/t2pks" in st_hmmpath:
            self._category = "t2pks"
        if "antismash/modules/sactipeptides" in st_hmmpath:
            self._category = "sactipeptides"
            if "antismash/modules/sactipeptides/data/non_biosyn_hmms" in st_hmmpath:
                self._subcategory = "non_biosyn_hmms"
        if "antismash/modules/rrefinder" in st_hmmpath:
            self._category = "rrefinder"
        if "antismash/modules/lassopeptides" in st_hmmpath:
            self._category = "lassopeptides"


class HmmParse:
    def __init__(self) -> None:
        self.models = dict()
        self.temp_model = HmmModel()
        self.dict_key_index = 0

    def read(self, filepath, base_dir=None):
        """
        Reads HMM models from file and adds them to the models dictionary.

        Args:
            filepath (str): The path to the file containing the HMM models.
            base_dir (str, optional): The base directory to use when resolving relative file paths and assigning the HMM source.

        Returns:
            None
        """
        for i, model in enumerate(
            self.read_model_generator(filepath=filepath, base_dir=base_dir)
        ):
            log.debug(f"Reading model {str(i + 1)} from {filepath}")
            model._n = self.dict_key_index
            self.models[self.dict_key_index] = model
            self.dict_key_index += 1

    def read_model_generator(self, filepath, base_dir=None):
        # this is used to read a single model
        # _add_line_contents is used to switch from attribute addition to HMM model additon by switching the function
        # to HMM model's functon after seeing "HMM ", to the end of the model
        # then switch back before the next model starts
        _add_line_contents = self.temp_model.add_attr
        with fh.open_read(filepath) as h:
            if not base_dir:
                base_dir = Path(filepath).parents[0]
            self.temp_model._base_dir = base_dir
            self.temp_model._abs_path = filepath
            self.temp_model._assign_relative_path()
            self.temp_model.assign_category()
            self.temp_model._model_source = Path(base_dir).name
            for line in h:
                if line.startswith("HMM "):
                    # switch to model addition
                    _add_line_contents = self.temp_model.add_model
                if line.startswith("//"):
                    # yield model
                    self.temp_model.add_model_hash()
                    self.temp_model.find_pfam_accessions()
                    yield self.temp_model
                    # Reset temp_model
                    self.temp_model = HmmModel()
                    self.temp_model._base_dir = base_dir
                    self.temp_model._abs_path = filepath
                    self.temp_model._assign_relative_path()
                    self.temp_model.assign_category()
                    self.temp_model._model_source = Path(base_dir).name
                    _add_line_contents = self.temp_model.add_attr
                    continue
                # add attributes or model
                _add_line_contents(line)

    def write_all(self, outpath, hash_as_name=False):
        # if only self.read(), then this should write exactly the same file back out
        # (input/output files will have the same hash)
        for i in self.models.values():
            _ = i.write(outpath, mode="w", hash_as_name=hash_as_name)

    def write_metadata_tsv(self, outdir: str, header: bool = False):
        """
        The function writes metadata for a collection of models to a TSV file

        Args:
          outdir (str): The `outdir` parameter is a string that specifies the directory where the output
        "all.hmminfo" TSV file will be written to.
          header (bool): The `header` parameter is a boolean flag that determines whether or not to
        write the header row in the TSV file. Defaults to False
        """
        with open(os.path.join(outdir, "all.hmminfo"), "w") as tsv_file:
            all_hmms_file_writer = csv.DictWriter(
                tsv_file, self.temp_model.all_attributes().keys(), delimiter="\t"
            )
            if header:
                all_hmms_file_writer.writeheader()
            for model in self.models.values():
                _ = all_hmms_file_writer.writerow(model.all_attributes())

    def write_hmm_node_tsv(self, outdir):
        """
        The function writes a list of HMM model hash values to a single-column TSV file.

        Args:
          outdir: The `outdir` parameter is the directory where the output "sg_hmm_nodes" file will be written.
        """
        with open(os.path.join(outdir, "sg_hmm_nodes"), "w") as h:
            hashgen = list({i._new_hash for i in self.models.values()})
            hashgen.sort()
            for i in hashgen:
                h.write(i)
                h.write("\n")


class HmmModelHandler(HmmParse):
    """
    The `HmmModelHandler` class is responsible for handling and manipulating Hidden Markov Model (HMM)
    models, including removing duplicate and outdated models, writing non-redundant HMM files, and
    managing metadata related to the models.
    """

    def __init__(
        self,
    ) -> None:
        super().__init__()
        self.cull_index = []
        self._pfam_removed = {}
        self._same_hash_removed = {}

    def get_all_model_hashes(self):
        """
        The function `get_all_model_hashes` returns a list of the `_hash` attribute values for all the
        models in the `self.models` dictionary.

        Returns:
          List of all the values of the "_hash" attribute for each item in the
        "models" dictionary.
        """
        return [getattr(i, "_hash") for i in self.models.values()]

    def hydrate_cull(self):
        self.cull_index = set(self.models.keys())

    def remove_duplicate_and_old_pfam(self):
        """
        The function removes duplicate and outdated Pfam models, updating metadata in Neo4j if
        necessary.
        """
        cull_at_start = len(self.cull_index)
        # create a dict with
        redundant_pfam = defaultdict(list)
        for k, v in self.models.items():
            if v._pfam_accession:
                redundant_pfam[v._pfam_accession].append(k)
        redundant_pfam = {k: v for k, v in redundant_pfam.items() if len(v) > 1}
        for pfam_id, model_index_list in redundant_pfam.items():
            z = {i: self.models.get(i)._pfam_version for i in model_index_list}
            max_ind = max(z, key=z.get)
            # remove the key of the max
            max_ind_version = z.pop(max_ind)
            deprecated_ind = z.keys()
            for model_index in deprecated_ind:
                # remove from culled list
                try:
                    self.cull_index.remove(model_index)
                except Exception:
                    pass
                # if the version is not the same as the highest version
                # then change metadata for Neo4j

                if self.models[model_index].ACC != self.models[max_ind].ACC:
                    # replace hash in deprecated models so the model metadata
                    # points to the max_ind model in Neo4j
                    self.models[model_index]._new_hash = self.models[max_ind]._hash
                    # change deprecated model ACC to reflect that an updated model was used
                    self.models[model_index]._notes.append(
                        f"{self.models[model_index].ACC}_updated_to_{self.models[max_ind].ACC}"
                    )
            self._pfam_removed[pfam_id] = {
                "kept": f"{pfam_id}.{max_ind_version}",
                "removed": [f"{pfam_id}.{i}" for i in z.values()],
            }
        log.info(
            f"Updated {cull_at_start - len(self.cull_index)} pfam models to a more recent version"
        )

    def remove_duplicate_hash(self):
        """
        The function removes duplicate hash values from a dictionary and updates the cull index and
        model notes accordingly.
        """
        cull_at_start = len(self.cull_index)
        redundant_hash = defaultdict(list)
        for k, v in self.models.items():
            redundant_hash[v._hash].append(k)
        redundant_hash = {k: v for k, v in redundant_hash.items() if len(v) > 1}
        for uid, model_index_list in redundant_hash.items():
            pass
            keep_this_one = model_index_list.pop(0)
            for i in model_index_list:
                # might have removed in another process (e.g. pfam deduplication), so check first
                if i in self.cull_index:
                    self.cull_index.remove(i)
                self.models[i]._notes.append(
                    f"{self.models[keep_this_one]._rel_path} was used instead of this model"
                )
            self._same_hash_removed[uid] = {
                "removed": model_index_list,
                "used": keep_this_one,
            }
        log.info(
            f"{cull_at_start - len(self.cull_index)} identical HMM models removed from culled list"
        )

    def write(
        self,
        outdir: str,
        hash_as_name: bool = False,
    ):
        """Write a non-redundant HMM file, optionally split into n-files, or split based on presence
        of cutoff values in a model

        Args:
            outdir (str): where to write the outfile
            n_files (int,deprecated): split the files into n-files (deprecated). Defaults to 1.
            hash_as_name (bool, optional): Replace model NAME field with the hash of model. Defaults to False.
        """
        n_files = 1
        if any(Path(outdir).glob("socialgene_nr_hmms*")):
            raise FileExistsError(
                f"Found existing socialgene_nr_hmms* file(s) in {outdir}; cowardly not continuing"
            )
        models_without_ga = [k for k, v in self.models.items() if not v.GA]
        models_without_ga = list(
            set(models_without_ga).intersection(set(self.cull_index))
        )
        models_with_ga = [k for k, v in self.models.items() if v.GA]
        models_with_ga = list(set(models_with_ga).intersection(set(self.cull_index)))
        file_counter = 1

        def hmm_filename_with_ga(file_counter, n_files):
            return (
                f"socialgene_nr_hmms_file_with_cutoffs_{file_counter}_of_{n_files}.hmm"
            )

        def hmm_filename_without_ga(file_counter, n_files):
            return f"socialgene_nr_hmms_file_without_cutoffs_{file_counter}_of_{n_files}.hmm"

        for model_index in models_without_ga:
            self.models[model_index].write(
                outpath=Path(outdir, hmm_filename_without_ga(file_counter, n_files)),
                mode="a",
                hash_as_name=hash_as_name,
            )
        for model_index in models_with_ga:
            self.models[model_index].write(
                outpath=Path(outdir, hmm_filename_with_ga(file_counter, n_files)),
                mode="a",
                hash_as_name=hash_as_name,
            )



import pandas as pd

# tax_url = "ftp://ftp.ncbi.nih.gov/pub/taxonomy/taxdump.tar.gz"
# tax_dump_hash_url = "ftp://ftp.ncbi.nih.gov/pub/taxonomy/taxdump.tar.gz.md5"


def process_names_dmp(file_handle):
    df = pd.read_csv(
        file_handle,
        sep="|",
        header=None,
        index_col=False,
        names=[
            "tax_id",
            "name_txt",
            "unique_name",
            "name_class",
        ],
        engine="c",
    )
    # remove extra \t
    df["name_txt"] = df["name_txt"].str.strip()
    df["unique_name"] = df["unique_name"].str.strip()
    df["name_class"] = df["name_class"].str.strip()
    # remove synonyms
    df = df[df["name_class"] == "scientific name"]
    df = df.iloc[:, [0, 1]]
    return df


def process_nodes_dmp(file_handle):
    df2 = pd.read_csv(
        file_handle,
        sep="|",
        header=None,
        index_col=False,
        names=[
            "tax_id",
            "parent_tax_id",
            "rank",
            "embl_code",
            "division_id",
            "inherited_div_flag",
            "genetic_code_id",
            "inherited_GC__flag",
            "mitochondrial_genetic_code_id",
            "inherited_MGC_flag",
            "GenBank_hidden_flag",
            "hidden_subtree_root_flag",
            "comments",
        ],
        engine="c",
    )
    df2["rank"] = df2["rank"].str.strip()
    df2["embl_code"] = df2["embl_code"].str.strip()
    df2["comments"] = df2["comments"].str.strip()
    return df2


def merge_taxonomy(names_df, nodes_df) -> pd.DataFrame:
    return names_df.merge(nodes_df.iloc[:, [0, 2]], how="inner", on="tax_id")
